-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
--Mandatory controller parameters
M.type = "main"
M.relevantDevice = nil
M.engineInfo = {}
M.gearboxInfo = {isReverse = false}
M.fireEngineTemperature = 0
M.throttle = 0
M.brake = 0
M.clutchRatio = 0

-----

local min = math.min
local max = math.max

local avToRPM = 9.549296596425384

local engineLeft = nil
local engineRight = nil
local propShaftLeft = nil
local propShaftRight = nil

local landingGearFinalState = "open"
local landingGearState = "open"
local landingGearTimer = 0

local bombDropCounter = 0

local rpmRange = 0
local idleRPM = 0
local rpmErrorIntegralLeft = 0
local rpmErrorIntegralRight = 0

local leftWheel = nil
local rightWheel = nil

local starterPressed = false
local starterPressedTime = 0

local energyStorageData = {
  tanks = {},
  ratio = 0,
  volume = 0,
  capacity = 0,
  invEnergyStorageCount = 0,
}

local flapsOff = 0
local flapsBattle = 0.2
local flapsTakeoff = 0.4
local flapsLanding = 1

local controlMode = nil
local maxElevator = 0
local maxAileron = 0
local maxRudder = 0

local function toggleFlaps()
  if electrics.values.flap > 0 then
    electrics.values.flap = flapsOff
  elseif electrics.values.gear_front_pressure == 0 and electrics.values.airspeed <= 60 then
    electrics.values.flap = flapsTakeoff
    gui.message("Takeoff flaps", 2, "flapmessage")
  elseif electrics.values.gear_front_pressure == 0 and electrics.values.airspeed > 60 then
    electrics.values.flap = flapsLanding
    gui.message("Landing flaps", 2, "flapmessage")
  elseif electrics.values.gear_front_pressure ~= 0 then
    electrics.values.flap = flapsBattle
    gui.message("Combat flaps", 2, "flapmessage")
  end
end

local function toggleLandingGear()
  if landingGearFinalState == "open" then
    landingGearState = "closing"
  elseif landingGearFinalState == "closed" then
    landingGearState = "opening"
  end
end

local function toggleBombDoors()
  electrics.values.bay_doors = 1 - electrics.values.bay_doors
end

local function dropBomb()
  if bombDropCounter < 4 and electrics.values.bay_doors == 1 then
    if bombDropCounter > 0 then
      electrics.values["bomb_release"..tostring(bombDropCounter)] = 0
    end
    bombDropCounter = bombDropCounter + 1
    electrics.values["bomb_release"..tostring(bombDropCounter)] = 1
  end
end

local function updateLandingGear(dt)
  if landingGearState == "opening" then
    landingGearState = "openingDoorsOpen"
    landingGearTimer = 0

  elseif landingGearState == "openingDoorsOpen" then
    electrics.values.gear_doors_pressure = 1
    electrics.values.gear_doors_pressure_2 = 0
    if landingGearTimer > 1 then
      landingGearState = "extendingGears"
      landingGearTimer = 0
    end

  elseif landingGearState == "extendingGears" then
    electrics.values.gear_front_pressure = 0
    electrics.values.gear_main_pressure = 0
    if landingGearTimer > 2.5 then
      landingGearState = "closingDoorsOpen"
      landingGearTimer = 0
    end

  elseif landingGearState == "closingDoorsOpen" then
    electrics.values.gear_doors_pressure = 0
    electrics.values.gear_doors_pressure_2 = 0
    if landingGearTimer > 1 then
      landingGearState = "open"
      landingGearFinalState = "open"
      landingGearTimer = 0
    end

  elseif landingGearState == "closing" then
    landingGearState = "openingDoorsClosed"
    landingGearTimer = 0

  elseif landingGearState == "openingDoorsClosed" then
    electrics.values.gear_doors_pressure = 1
    electrics.values.gear_doors_pressure_2 = 0
    if landingGearTimer > 1 then
      landingGearState = "retractingGears"
      landingGearTimer = 0
    end

  elseif landingGearState == "retractingGears" then
    electrics.values.gear_front_pressure = 1
    electrics.values.gear_main_pressure = 1
    if landingGearTimer > 2.5 then
      landingGearState = "closingDoorsClosed"
      landingGearTimer = 0
    end

  elseif landingGearState == "closingDoorsClosed" then
    electrics.values.gear_doors_pressure = 0
    electrics.values.gear_doors_pressure_2 = 1
    if landingGearTimer > 1 then
      landingGearState = "closed"
      landingGearFinalState = "closed"
      landingGearTimer = 0
    end
  end

  landingGearTimer = landingGearTimer + dt
end

local function updateGFX(dt)
  local throttle = electrics.values.throttleB25 or 0
  throttle = min(max(throttle + (electrics.values.throttleB25Increase * dt * 0.5) - (electrics.values.throttleB25Decrease * dt * 0.5), 0), 1)
  local airspeed = electrics.values.airspeed

  electrics.values.elevator = min(max(electrics.values.elevator, -maxElevator), maxElevator)
  electrics.values.aileron = min(max(electrics.values.aileron, -maxAileron), maxAileron)
  electrics.values.rudder = min(max(electrics.values.rudder, -maxRudder), maxRudder)

  --------DIFFERENTIAL BRAKES--------
  --add up the general brake with the steering brakes
  local rudderBrakeInput = electrics.values.airspeed < 30 and (input.rudder or 0) * 0.5 or 0
  local brake = input.brakeB25 or 0
  if throttle == 0 and airspeed < 5 then
    brake = 0.5
  end
  local brakeR = min(max(brake + rudderBrakeInput, 0), 1)
  local brakeL = min(max(brake - rudderBrakeInput, 0), 1)

  --scale the parking brake torque to match our desired torque
  if rightWheel then
    rightWheel.desiredBrakingTorque = rightWheel.brakeTorque * brakeR
  end
  if leftWheel then
    leftWheel.desiredBrakingTorque = leftWheel.brakeTorque * brakeL
  end

  local rpmLeft = engineLeft and engineLeft.outputAV1 * avToRPM or 0
  local rpmRight = engineRight and engineRight.outputAV1 * avToRPM or 0
  local rpm = max(rpmLeft, rpmRight)

  local targetRPM = rpmRange * throttle + idleRPM
  local rpmErrorLeft = rpmLeft - targetRPM
  local rpmErrorRight = rpmRight - targetRPM
  rpmErrorIntegralLeft = min(max(rpmErrorIntegralLeft - rpmErrorLeft * dt, 0), 5)
  rpmErrorIntegralRight = min(max(rpmErrorIntegralRight - rpmErrorRight * dt, 0), 5)

  local throttleLeft = min(max(rpmErrorIntegralLeft * 0.2, 0), 1)
  local throttleRight = min(max(rpmErrorIntegralRight * 0.2, 0), 1)

  if engineLeft then
    local stallPreventLeft = min(max((engineLeft.outputAV1 * 0.95 - engineLeft.idleAV * 0.35 ) / (engineLeft.idleAV * 0.1), 0), 1)
    electrics.values.clutchRatioLeft = (engineLeft.ignitionCoef < 1 or engineLeft.starterEngagedCoef >= 1) and 1 or stallPreventLeft * stallPreventLeft
  end

  if engineRight then
    local stallPreventRight = min(max((engineRight.outputAV1 * 0.95 - engineRight.idleAV * 0.35 ) / (engineRight.idleAV * 0.1), 0), 1)
  electrics.values.clutchRatioRight = (engineRight.ignitionCoef < 1 or engineRight.starterEngagedCoef >= 1) and 1 or stallPreventRight * stallPreventRight
  end

  local propPitchInput = electrics.values.propPitch
  local propPitchLeft =  max(propPitchInput, 0.3)
  local propPitchRight =  max(propPitchInput, 0.3)
  if propPitchInput < 0 then
    propPitchLeft = min(max(rpmLeft / 2800, 0), 1)
    propPitchRight = min(max(rpmRight / 2800, 0), 1)
  end

  if starterPressed then
    starterPressedTime = starterPressedTime + dt
  else
    starterPressedTime = 0
  end

  energyStorageData.ratio = 0
  energyStorageData.volume = 0

  for _,s in pairs(energyStorageData.tanks) do
    local energyStorage = energyStorage.getStorage(s)
    if energyStorage then
      energyStorageData.ratio = energyStorageData.ratio + energyStorage.remainingRatio
      energyStorageData.volume = energyStorageData.volume + energyStorage.remainingVolume
    end
  end
  energyStorageData.ratio = energyStorageData.ratio * energyStorageData.invEnergyStorageCount

  electrics.values.fuel = energyStorageData.ratio or 0
  electrics.values.lowfuel = electrics.values.fuel < 0.1
  electrics.values.fuelCapacity = energyStorageData.invEnergyStorageCount > 0 and (energyStorageData.capacity or 0) or 1
  electrics.values.fuelVolume = energyStorageData.volume

  electrics.values.throttle = 0
  electrics.values.brake = brake
  electrics.values.clutch = 0
  electrics.values.clutchRatio = 1
  electrics.values.gear = string.format("%d%%", throttle * 100)
  electrics.values.gearIndex = 0
  electrics.values.rpm = rpm
  electrics.values.rpmLeft = rpmLeft
  electrics.values.rpmRight = rpmRight
  electrics.values.throttleLeft = throttleLeft
  electrics.values.throttleRight = throttleRight
  electrics.values.wheelspeed = electrics.values.airspeed
  electrics.values.engineLoad = max((not engineLeft or engineLeft.isDisabled) and 0 or engineLeft.engineLoad, (not engineRight or engineRight.isDisabled) and 0 or engineRight.engineLoad)

  electrics.values.propPitchLeft = propPitchLeft
  electrics.values.propPitchRight = propPitchRight

  local bulletTimeCoef = bullettime.get()

  electrics.values.prop_blurr_left = (propShaftLeft and propShaftLeft.outputAV1 * bulletTimeCoef > 15) and propShaftLeft.outputAV1 * bulletTimeCoef / 150 or 0
  electrics.values.prop_blurr_right = (propShaftRight and propShaftRight.outputAV1 * bulletTimeCoef > 15) and propShaftRight.outputAV1 * bulletTimeCoef / 150 or 0

  updateLandingGear(dt)

  electrics.values.throttleB25 = throttle


  M.engineInfo = {
    rpm
    , engineLeft and engineLeft.maxAV * avToRPM or 0
    , 0
    , 0
    , electrics.values.rpm
    , electrics.values.gear
    , 0
    , 0
    , max(engineLeft and engineLeft.outputTorque1 or 0, engineRight and engineRight.outputTorque1 or 0)
    , max(propShaftLeft and propShaftLeft.outputTorque1 or 0, propShaftRight and propShaftRight.outputTorque1 or 0)
    , obj:getVelocity():length()  -- airspeed
    , electrics.values.fuelVolume
    , electrics.values.fuelCapacity
    , "manual"
    , obj:getID()
    , 0
    , 0
    , 1
    , electrics.values.engineLoad
  }
end

local function setStarter(enabled)
  starterPressed = enabled
  if engineLeft then
    engineLeft:setIgnition(enabled and 1 or 0)
    if engineLeft.isStalled and enabled then
      engineLeft:activateStarter()
    end
    if not enabled then
      if (engineLeft.outputAV1 > engineLeft.idleAV * 0.5 and engineLeft.starterEngagedCoef >= 1) or starterPressedTime > 0.5 then
        engineLeft:setIgnition(1)
      end
      engineLeft:deactivateStarter()
    end
  end

  if engineRight then
    engineRight:setIgnition(enabled and 1 or 0)
    if engineRight.isStalled and enabled then
      engineRight:activateStarter()
    end
    if not enabled then
      if (engineRight.outputAV1 > engineRight.idleAV * 0.5 and engineRight.starterEngagedCoef >= 1) or starterPressedTime > 0.5 then
        engineRight:setIgnition(1)
      end
      engineRight:deactivateStarter()
    end
  end
end

local function setEngineIgnition(value)
  if engineLeft then
    engineLeft:setIgnition(value and 1 or 0)
  end
  if engineRight then
    engineRight:setIgnition(value and 1 or 0)
  end
end

local function init(jbeamData)
  engineLeft = powertrain.getDevice("engineLeft")
  engineRight = powertrain.getDevice("engineRight")
  propShaftLeft = powertrain.getDevice("propShaftLeft")
  propShaftRight = powertrain.getDevice("propShaftRight")

  idleRPM = engineLeft and engineLeft.idleAV * avToRPM or 0
  rpmRange = engineLeft and engineLeft.maxRPM - idleRPM or 0

  landingGearFinalState = "open"
  landingGearState = "open"
  landingGearTimer = 0

  bombDropCounter = 0
  starterPressed = false
  starterPressedTime = 0

  for _,wd in pairs(wheels.wheels) do
    if wd.name == "Wheel_R" then rightWheel = wd end
    if wd.name == "Wheel_L" then leftWheel = wd end
  end

  electrics.values.brake_combined = 0
  electrics.values.brake_r = 0
  electrics.values.brake_l = 0

  electrics.values.throttleB25 = 0
  electrics.values.throttleB25Increase = 0
  electrics.values.throttleB25Decrease = 0

  electrics.values.gear_front_pressure = 0
  electrics.values.gear_main_pressure = 0
  electrics.values.gear_doors_pressure = 0
  electrics.values.gear_doors_pressure_2 = 0

  electrics.values.bay_doors = 0
  electrics.values.bomb_release1 = 0
  electrics.values.bomb_release2 = 0
  electrics.values.bomb_release3 = 0
  electrics.values.bomb_release4 = 0

  electrics.values.prop_blurr_left = 0.0
  electrics.values.prop_blurr_right = 0.0

  electrics.values.propPitch = -1

  electrics.values.propPitchLeft = 0
  electrics.values.propPitchRight = 0

  energyStorageData = {
    tanks = {},
    ratio = 0,
    volume = 0,
    capacity = 0,
    invEnergyStorageCount = 0,
  }

  if engineLeft then
    for _,s in pairs(engineLeft.energyStorage or {}) do
      table.insert(energyStorageData.tanks, s)
    end
  end
  if engineRight then
    for _,s in pairs(engineRight.energyStorage or {}) do
      table.insert(energyStorageData.tanks, s)
    end
  end

  local energyStorageCount = 0
  for _,s in ipairs(energyStorageData.tanks) do
    local energyStorage = energyStorage.getStorage(s)
    if energyStorage then
      energyStorageData.capacity = energyStorageData.capacity + energyStorage.capacity
      energyStorageCount = energyStorageCount + 1
    end
  end

  energyStorageData.invEnergyStorageCount = energyStorageCount > 0 and 1 / energyStorageCount or 0

  controlMode = jbeamData.controlMode or "realistic"
  maxElevator = jbeamData.maxElevator or 1
  maxAileron = jbeamData.maxAileron or 1
  maxRudder = jbeamData.maxRudder or 1

  mapmgr.enableTracking()
  obj:queueGameEngineLua(string.format("be:getObjectByID(%d):setField('renderDistance', '', '%d')", obj:getID(), 7500))
end

local function sendTorqueData()
  if not playerInfo.firstPlayerSeated then return end
  if engineLeft then
    engineLeft:sendTorqueData()
  end
  if engineRight then
    engineRight:sendTorqueData()
  end
end

M.init = init
M.updateGFX = updateGFX

--Mandatory main controller API
M.shiftUp = nop
M.shiftDown = nop
M.shiftToGearIndex = nop
M.cycleGearboxModes = nop
M.setStarter = setStarter
M.freeze = nop
M.sendTorqueData = sendTorqueData
-------------------------------
M.setEngineIgnition = setEngineIgnition

M.toggleLandingGear = toggleLandingGear
M.toggleFlaps = toggleFlaps
M.toggleBombDoors = toggleBombDoors
M.dropBomb = dropBomb

return M